//
// /build/openvkl/src/build/openvkl/devices/cpu/ispc_openvkl_module_cpu_device_8_avxavx2/VdbVolume_ispc.h
// (Header automatically generated by the ispc compiler.)
// DO NOT EDIT THIS FILE.
//

#pragma once
#include <stdint.h>



#ifdef __cplusplus
namespace ispc { /* namespace */
#endif // __cplusplus

///////////////////////////////////////////////////////////////////////////
// Enumerator types with external visibility from ispc code
///////////////////////////////////////////////////////////////////////////

#ifndef __ISPC_ENUM_VKLTemporalFormat__
#define __ISPC_ENUM_VKLTemporalFormat__
enum VKLTemporalFormat {
    VKL_TEMPORAL_FORMAT_CONSTANT = 0,
    VKL_TEMPORAL_FORMAT_STRUCTURED = 1,
    VKL_TEMPORAL_FORMAT_UNSTRUCTURED = 2,
    VKL_TEMPORAL_FORMAT_INVALID = 100 
};
#endif


/* Portable alignment macro that works across different compilers and standards */
#if defined(__cplusplus) && __cplusplus >= 201103L
/* C++11 or newer - use alignas keyword */
#define __ISPC_ALIGN__(x) alignas(x)
#elif defined(__GNUC__) || defined(__clang__)
/* GCC or Clang - use __attribute__ */
#define __ISPC_ALIGN__(x) __attribute__((aligned(x)))
#elif defined(_MSC_VER)
/* Microsoft Visual C++ - use __declspec */
#define __ISPC_ALIGN__(x) __declspec(align(x))
#else
/* Unknown compiler/standard - alignment not supported */
#define __ISPC_ALIGN__(x)
#warning "Alignment not supported on this compiler"
#endif // defined(__cplusplus) && __cplusplus >= 201103L
#ifndef __ISPC_ALIGNED_STRUCT__
#if defined(__clang__) || !defined(_MSC_VER) || _MSC_VER > 1943
// Clang, GCC, ICC, Visual Studio
#define __ISPC_ALIGNED_STRUCT__(s) struct __ISPC_ALIGN__(s)
#else
// Older Visual Studio
#define __ISPC_ALIGNED_STRUCT__(s) __ISPC_ALIGN__(s) struct
#endif // defined(__clang__) || !defined(_MSC_VER) || _MSC_VER > 1943
#endif // __ISPC_ALIGNED_STRUCT__

#ifndef __ISPC_STRUCT_vec3i__
#define __ISPC_STRUCT_vec3i__
struct vec3i {
    int32_t x;
    int32_t y;
    int32_t z;
};
#endif

#ifndef __ISPC_STRUCT_box1f__
#define __ISPC_STRUCT_box1f__
struct box1f {
    float lower;
    float upper;
};
#endif

#ifndef __ISPC_STRUCT_vec3ui__
#define __ISPC_STRUCT_vec3ui__
struct vec3ui {
    uint32_t x;
    uint32_t y;
    uint32_t z;
};
#endif

#ifndef __ISPC_STRUCT_vec3f__
#define __ISPC_STRUCT_vec3f__
struct vec3f {
    float x;
    float y;
    float z;
};
#endif

#ifndef __ISPC_STRUCT_box3f__
#define __ISPC_STRUCT_box3f__
struct box3f {
    struct vec3f lower;
    struct vec3f upper;
};
#endif

#ifndef __ISPC_STRUCT_Data1D__
#define __ISPC_STRUCT_Data1D__
struct Data1D {
    const uint8_t * addr;
    uint64_t byteStride;
    uint64_t numItems;
    uint32_t dataType;
    bool compact;
};
#endif

#ifndef __ISPC_STRUCT_VdbLevel__
#define __ISPC_STRUCT_VdbLevel__
struct VdbLevel {
    uint64_t numNodes;
    struct vec3ui * origin;
    uint64_t * voxels;
    struct box1f * valueRange;
};
#endif

#ifndef __ISPC_STRUCT_VdbGrid__
#define __ISPC_STRUCT_VdbGrid__
struct VdbGrid {
    float objectToIndex[12];
    float indexToObject[12];
    struct vec3i rootOrigin;
    struct vec3ui activeSize;
    struct box3f domainBoundingBox;
    bool constantCellData;
    uint64_t numLeaves;
    uint32_t numAttributes;
    bool allLeavesCompact;
    bool allLeavesConstant;
    uint32_t * attributeTypes;
    const int32_t * leafStructuredTimesteps;
    struct Data1D * leafUnstructuredIndices;
    struct Data1D * leafUnstructuredTimes;
    struct Data1D * leafData;
    bool packedAddressing32;
    struct Data1D * nodesPackedDense;
    struct Data1D * nodesPackedTile;
    bool dense;
    struct vec3i denseDimensions;
    struct Data1D * denseData;
    enum VKLTemporalFormat denseTemporalFormat;
    int32_t denseTemporallyStructuredNumTimesteps;
    struct Data1D denseTemporallyUnstructuredIndices;
    struct Data1D denseTemporallyUnstructuredTimes;
    struct VdbLevel levels[3];
};
#endif


///////////////////////////////////////////////////////////////////////////
// Functions exported from ispc code
///////////////////////////////////////////////////////////////////////////
#if defined(__cplusplus) && (! defined(__ISPC_NO_EXTERN_C) || !__ISPC_NO_EXTERN_C )
extern "C" {
#endif // __cplusplus
    extern void VdbVolume_Constructor8(void * _self);
#if defined(__cplusplus)
    extern void VdbVolume__declare_types_dummy8(struct vec3i &dummy_vec3i, struct box1f &dummy_range1f);
#else
    extern void VdbVolume__declare_types_dummy8(struct vec3i *dummy_vec3i, struct box1f *dummy_range1f);
#endif // VdbVolume__declare_types_dummy8 function declaraion
    extern void VdbVolume_destroy8(void * _self);
    extern void VdbVolume_setGrid8(void * _self, const struct VdbGrid * grid);
#if defined(__cplusplus) && (! defined(__ISPC_NO_EXTERN_C) || !__ISPC_NO_EXTERN_C )
} /* end extern C */
#endif // __cplusplus



#ifdef __cplusplus
} /* namespace */
#endif // __cplusplus
